/*
 * UndoRedoEdit.java
 *
 * Created on April 18, 2005, 9:44 PM
 *
 Copyright (c) 2005 Forward Computing and Control Pty. Ltd.
NSW. Australia,  www.forward.com.au
All rights reserved.
 
Redistribution of the source of this package, with or without
modification, is NOT permitted.
 
Redistribution of the compiled java .class files of this package, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions in compiled java .class form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.
 
2. If the software is modified, the compiled java .class distribution must
   include, in the documentation and/or other materials provided with the
   distribution, a statement that the software has been modified and
   identify the person who was responsable for the modifications.
 
3. Neither the name of the author nor Forward Computing and Control may be
   used to endorse or promote products derived from this software
   without specific prior written permission.
 
4.  All advertising materials mentioning features or use of this
    software must display the following acknowledgment:
    "This product includes software developed by Matthew Ford
     and Forward Computing and Control Pty. Ltd.
    (http://www.forward.com.au)."
 
 
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package au.com.forward.undoRedo;

import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;

/**
 *
 * @author Matthew Ford
 */
public abstract class UndoRedoEdit extends AbstractUndoableEdit implements ICanBeRolledBack, IHaveDescription {
  
  private Long frameMarker=null;
  protected String description = "";
  
  
  /** Creates a new instance of UndoRedoEdit */
  public UndoRedoEdit() {
  }
  
  /**
   * Gets the Description of the UndoableEdit.
   * <br>If getting the description will take some time,
   * such as generating a diff for example. Then
   * display a short message and start a thread to call back
   * to the textDisplay.setText() when the description is ready.
   * <br>If the selection has changed before textDisplay.setText() is
   * called, then the text is discarded.
   * <br>{@link #cancelGetDescription()} is called when the selection
   * changes and the text is no longer needed.
   */
  public void getDescription(IDisplayText textDisplay) {
    textDisplay.setText(description);
  }
  
  /**
   * Cancels getDescription
   * <br>This is called when the undo/redo tree selection changes
   * to cancel the previous getDescription.
   */
  public void cancelGetDescription() {
    // do nothing
  }
  
  /**
   * Called by the UndoRedoManager to set
   * the frame marker for future roll back.
   * <br>This method is called when the UndoableEdit is
   * registered with the UndoRedoManager and each time an
   * undo or redo operation is performed so it needs to be
   * saved with the task performing the operation.
   *
   * @param marker the frameMarker associated with this operation
   */
  public void setFrameMarker(Long marker) {
    this.frameMarker = marker;
  }
  
  /**
   * Returns the last frameMarker set
   * <br>Because the frameMarker is set for the
   * original edit and every subsequent undo/redo
   * of it, the frameMarker needs to be
   * saved with the task preforming the operation.
   * <p>
   * This implementation nulls the frameMarker when
   * it is returned. This means this method can only be
   * called once for each marker set.  Nulling the
   * frameMarker here allows it to be garbaged collected
   * from the UndoRedoManager when there are not other
   * references.
   *
   * @return the last frameMarker set.
   */
  public Long getFrameMarker() {
    Long rtn = frameMarker;
    frameMarker = null;
    return rtn;
  }
  
  /**
   * Sets the undo or redo state.
   * <br>Called multiple times by the UndoRedoManager when rolling back to
   * reset the UndoableEdit to its initial state.
   * <p>
   * This method should also do any other clean up necessary
   * in the UndoableEdit to return it to its previous state.
   * <p>This method is only called to toggle the state. That is
   * if the current state is canRedo() the only call will be
   * setUndoRedoState(true);
   *
   * @param undoState true to return UndoableEdit to undo state,
   * false to return it to a redo state.
   */
  public void setUndoState(boolean undoState) {
    if (undoState) {
      try {
        super.redo();
      } catch (CannotUndoException cex) {
        cex.printStackTrace();
        // ignore
      }
    } else {
      try {
        super.undo();
      } catch (CannotRedoException cex) {
        cex.printStackTrace();
        // ignore
      }
    }
  }
  
}
