/*
 * UndoRedoManager.java
 * Created on March 3, 2005, 8:36 PM by Matthew Ford
 *
Copyright (c) 2005 Forward Computing and Control Pty. Ltd.
NSW. Australia,  www.forward.com.au
All rights reserved.
 
Redistribution of the source of this package, with or without
modification, is NOT permitted.
 
Redistribution of the compiled java .class files of this package, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions in compiled java .class form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.
 
2. If the software is modified, the compiled java .class distribution must
   include, in the documentation and/or other materials provided with the
   distribution, a statement that the software has been modified and
   identify the person who was responsable for the modifications.
 
3. Neither the name of the author nor Forward Computing and Control may be
   used to endorse or promote products derived from this software
   without specific prior written permission.
 
4.  All advertising materials mentioning features or use of this
    software must display the following acknowledgment:
    "This product includes software developed by Matthew Ford
     and Forward Computing and Control Pty. Ltd.
    (http://www.forward.com.au)."
 
 
THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

package au.com.forward.undoRedo;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.util.Enumeration;
import java.util.Vector;
import java.util.WeakHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.JViewport;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoableEdit;


/**
 * This is the main class, it handles a tree of undo/redos.
 * The calls that modify the data structure, {@link #undo()},
 * {@link #redo()}, {@link #processUndoRedoCommand()} and {@link #undoableEditHappened(UndoableEditEvent)}
 * are not synchronized but are protected by a busy flag.  If the manager is already busy processing one of
 * these calls, another call to any of these will just quietly return doing nothing.
 * <br>If the manager is busy, calls to {@link #canUndo()} and {@link #canRedo()} will return false.
 * <br>This makes the UndoRedoManager data structure threadsafe and deadlock safe, but it is upto you to prevent
 * concurrent undoableEdits, undos and redos in order to keep the undo/redo data structure complete.
 *
 * @author Matthew Ford
 */
public class UndoRedoManager extends JSplitPane implements UndoableEditListener{
  
  /**
   * UndoRedoManager Version
   */
  public static final String VERSION="1.2";
  
  /**
   * The undoRedo JTree
   */
  protected JTree tree = null;
  
  private AtomicBoolean busy = new AtomicBoolean();
  private boolean blocked = false;
  private AtomicLong frameCounter = new AtomicLong(0);

  /**
   * The rootNode of the tree model.
   */
  protected UndoRedoTreeNode rootNode = null;
  
  /**
   * The last cloned rootNode. Can be null if
   *  last UndoableEdit did not implement 
   *  ICanBeRollback
   */
  private UndoRedoTreeNode lastClonedTree = null;

  /** 
   * True if this tree can be rolled back.
   * That is all registered undoable edits inplement
   * ICanBeRolledBack
   */
  private boolean hasRollBackSupport = false;
  
  /**
   * The undo/redo tree model
   */
  protected DefaultTreeModel treeModel = null;
  /**
   * A map holding weak keys to the tree frames.
   */
  protected WeakHashMap<Long,Frame> frameMarkerMap = new WeakHashMap<Long,Frame>();
  
  /**
   * The preferred width of the tree node text displays.
   */
  protected int preferredWidth = 50;
  
  private IDisplayText textDisplayComponent;
  /**
   * The scrollPane holding the extended description text component.
   */
  protected JScrollPane textScrollPane;
  /**
   * The scrollPane holding the undo/redo Tree.
   */
  protected JScrollPane treeScrollPane;
  
  /**
   * Creates a new instance of UndoRedoManager.
   *<br> Using the default Enter key action which undoes/redoes to the selected tree node.
   */
  public UndoRedoManager() {
    this(null,null,false);
  }
  
  
  /**
   * Creates a new instance of UndoRedoManager.
   * @param textComponent - the Component used to display the extended descriptions, 
   * <br>if null a default class extending JTextArea is used.
   * @param enter - the action to attach to the Enter key for the selected treenode,
   * <br> if null use a default action which undoes/redoes to
   * the selected node
   * @param canRollBack - true if this UndoRedoManager supports rollback. If true
   * then all UndoableEdits passed to {@link #undoableEditHappened } must
   * implement ICanBeRolledBack
   */
  public UndoRedoManager(Action enter,IDisplayText textComponent, boolean canRollBack) {
    super(JSplitPane. HORIZONTAL_SPLIT);
    hasRollBackSupport = canRollBack;
    if (enter == null) {
      enter = enterAction;
    }
    if (textComponent == null) {
      textComponent = new DisplayTextArea();
    }
    if (!(textComponent instanceof Component)) {
      throw new IllegalArgumentException("textComponent must extend from Component.");
    }
    textDisplayComponent = textComponent;
    
    rootNode = new UndoRedoTreeNode("Undo/Redo");
    treeModel = new DefaultTreeModel(rootNode);
    lastClonedTree = rootNode;
    
    tree = new JTree(treeModel);
    tree.setEditable(true);
    tree.getSelectionModel().setSelectionMode
            (TreeSelectionModel.SINGLE_TREE_SELECTION);
    tree.setShowsRootHandles(true);
    tree.setRootVisible(false);
    tree.setCellRenderer(new WidthEnabledCellRenderer());
    tree.setEditable(false);
    tree.addTreeSelectionListener(new UndoRedoTreeSelectionListener());
    //super.setViewportView(tree);
    setKeyboardMappings(enter);
    treeScrollPane = new JScrollPane(tree);
 //   treeScrollPane.getViewport().setScrollMode(JViewport.SIMPLE_SCROLL_MODE);
    textScrollPane = new JScrollPane((Component)getTextDisplayComponent());
    super.setLeftComponent(treeScrollPane);
    super.setRightComponent(textScrollPane);
    super.setResizeWeight(0.5);  // most of the display is for the tree
  }

  /**
   * Returns the rollback setting for the UndoRedoManager.
   * @return - true if rollback is supported, else false. <br>
   * It true then all UndoableEdits passed this manager must implement 
   * ICanBeRolledBack.
   */
  public boolean supportsRollBack() {
    return hasRollBackSupport;
  }
  
  /**
   * An undoable edit happened, handle it.
   * <br>If the the manager is already busy just return quietly.
   * <br>Wraps the UndoableEditEvent in an UndoRedo object and add it to the tree.
   * <br>Sets busy while it is processing.
   * @param e - the UndoableEditEvent
   */
  public void undoableEditHappened(UndoableEditEvent e) {
    if (!busy.compareAndSet(false,true) ) {
      // already busy
      return;
    }
    try {
      // wrap in an undoRedo obj
      UndoableEdit undoableEdit = e.getEdit();

      // do we need to take a snapshot of the current frame
      if ((!(undoableEdit instanceof ICanBeRolledBack)) && supportsRollBack()) {
          throw new RuntimeException("Tried to add UndoableEdit that does " +
                  "not implement ICanBeRolledBack to an UndoRedoManager that supports rollback.");
      }
      UndoRedo undoRedo = new UndoRedo(undoableEdit);
      addUndoRedo(undoRedo,true);
      
      if (supportsRollBack()) {
        ICanBeRolledBack cancellableEdit = (ICanBeRolledBack)undoableEdit;
        Frame frame = new Frame(getRoot());
        // save the frame and pass the reference to the undoableEdit
        Long key = new Long(frameCounter.getAndIncrement());
        frameMarkerMap.put(key,frame);
        //System.out.println("put key:"+key+"\nframe:"+frame);
        cancellableEdit.setFrameMarker(key);
        // now add edit
      }
      
    } finally {
      busy.set(false);
    }
  }
  

  /**
   * Rolls back undo/redo to the last sucessfull command.
   * <br>Use this to  recover the undo/redo stack to match
   * the last command that the server actually executed.
   * <p>
   * The frameMarker points to the frame prior to the
   * command being executed. The undo/redo is rolled back
   * to the following frame in the stack, i.e. the frame
   * after the command has been executed.
   * <br>The frameMarker is disposed of after this call.
   * <br>If the frameMarker is not valid (cannot be found)
   * just ignore quietly.  This allows this method to
   * be called in any order of a sequence of roll backs.
   * The oldest frameMarker will take precedence.
   *
   * @param frameMarkerKey  the marker before the one to roll back to.
   */
  public void syncToLastSuccessfull(Long frameMarkerKey) {
    // TODO what about busy need to sync here to protect tree.
//    System.out.println("called rollBackToAfter with "+frameMarkerKey);
    Frame frame = frameMarkerMap.get(frameMarkerKey);
//    cancelRollBack(frameMarkerKey); 
    if (frame == null) {
      // not other frame after this one so leave stack as is
      throw new RuntimeException("Could not sync to last successfull action. Frame was null");
    } else {
      // update treeModel to previous frame
      treeModel.setRoot(frame.getPostTreeRootNode());
    }
    
  }
  
  /**
   * Rolls back undo/redo to the last command that failed.
   * <br>Use this to  recover the undo/redo stack to match
   * the last command that the server actually executed.
   * <p>
   * Rolls back an undo/redo if it fails or is cancelled.
   * <br>This and following frameMarkers are disposed of after this call.
   * <br>If the frameMarker is not valid (cannot be found)
   * just ignore quietly.  This allows this method to
   * be called in any order of a sequence of roll backs.
   * The oldest frameMarker will take precedence.
   *
   * @param frameMarkerKey  the marker to roll back to.
   */
  public void cancel(Long frameMarkerKey) {
    // what about busy need to sync here to protect tree.
//    System.out.println("called rollBack with "+frameMarkerKey);
    Frame frame = frameMarkerMap.get(frameMarkerKey);
    // System.out.println("got frame "+frame);
//    cancelRollBack(frameMarkerKey); 
    if (frame == null) {
      // not other frame after this one so leave stack as is
      throw new RuntimeException("Could not cancel action. Frame was null");
    } else {
      // update treeModel to previous frame
      treeModel.setRoot(frame.getPreTreeRootNode());
    }
    
  }

  /**
   * Frees the resources associated with  any following frameMarker.
   * <br>Call after an UndoableEdit redo/undo is cancelled.
   * <br>Actually only the resources associated with frameMarkers after
   * the next one are released. The next frame marker is kept for possiable use by
   * {@link#rollBackToAfter()}
   * <p>FrameMarkers are only weekly held by
   * the UndoRedoManager so they can be garbage collected
   * when there are no other references.  Since the
   * marker is set in the UndoableEdit implementing this
   * interface, you need to null it out that reference,
   * and any others, in order have it garbage collected,
   * if you do not call this method.
   * <br>Any future calls to rollBack with this marker
   * will be quietly ignored.
   *
   * @param frameMarkerKey the marker to free
   */
//  private void cancelRollBack(Long frameMarkerKey) {
//    Long frameCounter = frameMarkerKey;
//    //System.out.println("called cancelRollBack with "+frameCounter);
//    // keep current maker
//    frameCounter++;
////    System.out.println(" removing  "+frameCounter);
//    Frame frame = frameMarkerMap.remove(frameCounter);
////    if (frame != null) {
////      System.out.println(" removed "+frameCounter);
////    }
//    while (frame != null) {
//      // remove all the later ones as we are rolling back to here
//      // this assumes later undo/redos are cancelled when this one is
//      frameCounter++;
//      frame = frameMarkerMap.remove(frameCounter);
////      if (frame != null) {
////        System.out.println(" removed "+frameCounter);
////      }
//    }
//  }

  
  
  /**
   * Frees the resources associated with a frameMarker.
   * <br>Call after an UndoableEdit redo/undo completes
   * successfully.
   * <p>FrameMarkers are only weekly held by
   * the UndoRedoManager so they can be garbage collected
   * when there are no other references.  Since the
   * marker is set in the UndoableEdit implementing this
   * interface, you need to null it out that reference,
   * and any others, in order have it garbage collected,
   * if you do not call this method.
   *
   * @param frameMarkerKey the marker to free
   */
  public void disposeOfRollBack(Long frameMarkerKey) {
    Long frameCounter = frameMarkerKey;
    //System.out.println("called disposeOfRollBack with "+frameCounter);
    //System.out.println(" removing  "+frameCounter);
    Frame frame = frameMarkerMap.remove(frameCounter);
//    if (frame != null) {
//      System.out.println(" removed "+frameCounter);
//    }
  }
  
  /**
   * Get the JTree to display
   * @return the tree showing the undo/redo branches
   */
  public JTree getTree() {
    return tree;
  }
  
  /**
   * Is the manager busy doing something.
   * @return true if UndoRedoManager is busy handling {@link #undoableEditHappened(UndoableEditEvent)}, {@link #undo()} or
   * {@link #redo()} or {@link #processUndoRedoCommand()}, else false
   */
  public boolean isBusy() {
    return busy.get();
  }
  
  /**
   * SetBlocked state in UndoRedoManager.
   * <br>Set to true to prevent undo, redo, and processUndoRedoCommand commands being processed.
   * <br>Note: Setting blocked true does not prevent undoableEditHappened() 
   * from processing new undoable events.  It only blocks user inputs.
   * @param flag  true to prevent undo, redo, and processUndoRedoCommand
   * from responding
   * <br>false to enable processing again.
   */
  public void setBlocked(boolean flag) {
    blocked = flag;  
  }
  
  
  /**
   * Is UndoRedoManager currently blocked from accepting undo,redo or processUndoRedoCommand commands
   * <br>Note: isBlocked() == true does not prevent undoableEditHappened() 
   * from processing new undoable events.  It only blocks user inputs.
   * @return true if undo, redo, and processUndoRedoCommand are blocked
   */
  public boolean isBlocked() {
    return blocked;
  }
  
  /**
   * Is UndoRedoManager busy or blocked?
   * @return true if either busy processing or blocked
   */
  public boolean isBusyOrBlocked() {
    return busy.get() || blocked;
  }
  
  /**
   * Is there an undo node in the main trunk that can be undone, does not check for dead nodes.
   *
   * @return true if there is an undo node, false if busy or all main branch nodes have been undone
   */
  public boolean canUndo() {
    if (!busy.compareAndSet(false,true) ) {
      // already busy
      return false;
    }
    try {
      return getUndoRedoPath(new Vector<UndoRedoTreeNode>(), true);
    } finally {
      busy.set(false);
    }
  }
  
  /**
   * Is there a redo branch at the top that can be redone, does not check for dead nodes.
   *
   * @return true if top node in a redo branch has not already been redone
   * after cancelling out the undo nodes with their redos
   * in the highest most redo branchs (depth first).
   * <br>False if busy or top node after cancellation is an undo or
   * all redo's nodes in the highest most redo branchs (depth first) have been redone.
   */
  public boolean canRedo() {
    if (!busy.compareAndSet(false,true) ) {
      // already busy
      return false;
    }
    try {
      return getUndoRedoPath(new Vector<UndoRedoTreeNode>(), false);
    } finally {
      busy.set(false);
    }
  }
  
  /**
   * Find the top undo and undo it.
   * <br>Just return if nothing to do or if the manager is busy.
   * <br>Sets busy while it is processing.
   * @throws javax.swing.undo.CannotUndoException  if cannot be undone.
   */
  public void undo() throws CannotUndoException {
    if (blocked) {
      return;
    }
    //  get the children
    if (!busy.compareAndSet(false,true) ) {
      // already busy
      return;
    }
    try {
      Vector<UndoRedoTreeNode> fullPath = new Vector<UndoRedoTreeNode>();
      if (getUndoRedoPath(fullPath, true)) {
        undoRedoTreePath(fullPath);
      }
    } finally {
      busy.set(false);
    }
  }
  
  /**
   * Find the top redo and redo it.
   * <br>Just return if nothing to do or the manager is busy.
   * <br>Sets busy while it is processing.
   * @throws javax.swing.undo.CannotRedoException  if cannot be redone.
   */
  public void redo() throws CannotRedoException {
    if (blocked) {
      return;
    }
    //  get the children
    if (!busy.compareAndSet(false,true) ) {
      // already busy
      return;
    }
    try {
      Vector<UndoRedoTreeNode> fullPath = new Vector<UndoRedoTreeNode>();
      if (getUndoRedoPath(fullPath, false)) {
        undoRedoTreePath(fullPath);
      }
    } finally {
      busy.set(false);
    }
  }
  
  
  /**
   * Undo/redo to the selected tree node.
   * <br>If no node selected or if already busy, just return quietly.
   * <br>Sets busy while it is processing.
   * <br>Note: if the undo/redo will take some time you should start another thread to run this method.
   *
   */
  public void processUndoRedoCommand()  {
    if (blocked) {
      return;
    }
    if (!busy.compareAndSet(false,true) ) {
      // already busy
      return;
    }
    try {
      TreePath treePath = tree.getSelectionPath();
      if (treePath == null) {
        return; // nothing selected
      }
      
      UndoRedoTreeNode selectedNode = (UndoRedoTreeNode)treePath.getLastPathComponent();
      if (selectedNode instanceof BranchDefaultMutableTreeNode) {
          // update treepath to deepest node
          selectedNode = ((BranchDefaultMutableTreeNode)selectedNode).getDeepestNode();
      }
      
      Object[] path = selectedNode.getPath();;
      Vector<UndoRedoTreeNode> fullPath = new Vector<UndoRedoTreeNode>();
      for (int i=0; i<path.length; i++) {
        // for each level add all the children above this node
        TreeNode path_i = (TreeNode)path[i];
        TreeNode parent = ((TreeNode)path_i).getParent();
        if (parent == null) {
          continue; // skip root
        }
        boolean notFound = true;
        for (int j=0; j<parent.getChildCount() && notFound; j++) {
          TreeNode child = (TreeNode)parent.getChildAt(j);
          if (child == path[i]) {
            notFound = false;
          } else {
            fullPath.add((UndoRedoTreeNode)child);
          }
        }
        fullPath.add((UndoRedoTreeNode)path_i);
      }
      fullPath = compressFullPath(fullPath);
      undoRedoTreePath(fullPath);
    } finally {
      busy.set(false); // finished
    }
  }
  
  /**
   * Sets the preferred width of the node display.
   * <br>Use this to prevent node descriptions being truncated.
   *
   * @param width the width to set, if it is <20 the width is set at 20
   */
  public void setPreferredDisplayWidth(int width) {
    if (width<20) {
      width = 20;
    }
    preferredWidth = width;
  }
  
  /**
   * Compare two DefaultMutableTreeNodes for UndoRedo complements.
   * <br>UndoRedo complements have the same UndoableEdit but
   * <br>undoRedo1.isUndo() == (! undoRedo2.isUndo())
   *<p>
   * Note: two BranchUndoRedo userObjects will compare as equal
   *
   * @param obj1 - node 1
   * @param obj2 - node 2
   * @return  true if node 1 is an UndoRedo complement of node 2, else false
   */
  public static boolean equalsComplement(UndoRedoTreeNode obj1, UndoRedoTreeNode obj2) {
    Object user1 = obj1.getUserObject();
    Object user2 = obj2.getUserObject();
    if ((user1 == null) || (!(user1 instanceof UndoRedo))) {
      return false;
    }
    return ((UndoRedo)user1).equalsComplement(user2);
  }
  
  
  
  /**
   * Returns the root node of the tree
   * @return root node
   */
  protected UndoRedoTreeNode getRoot() {
    return (UndoRedoTreeNode)treeModel.getRoot();
  }
  
  
  
  
  /**
   * Set the Enter keyboard mapping for the tree to the given Action.
   *
   * @param enter the action to use.
   */
  protected void setKeyboardMappings(Action enter) {
    InputMap im = new InputMap();
    im.setParent(tree.getInputMap());
    im.put(KeyStroke.getKeyStroke("ENTER"),enter.getValue(Action.NAME));
    im.put(KeyStroke.getKeyStroke("R"),redrawAction.getValue(Action.NAME));
    tree.setInputMap(WHEN_FOCUSED,im);
    
    ActionMap am = new ActionMap();
    am.setParent(tree.getActionMap());
    am.put(enter.getValue(Action.NAME),enter);
    am.put(redrawAction.getValue(Action.NAME),redrawAction);
    tree.setActionMap(am);
    
  }

  /**
   * Default enterAction which undoes/redoes to the selected node.
   */
  private Action redrawAction =
          new AbstractAction("redrawAction") {
    public void actionPerformed(ActionEvent evt) {
        UndoRedoTreeNode newRoot = (UndoRedoTreeNode)getRoot().clone();
//        treeModel = new DefaultTreeModel(rootNode);
//    TreeModel treeModel = tree.getModel();
//       tree.setModel(treeModel);
//      tree.invalidate();  
        treeModel.setRoot(getRoot());
//      UndoRedoManager.this.repaint();
//      treeScrollPane.setViewportView(tree);
//      UndoRedoManager.this.repaint();
System.out.println("redraw");      
    }
  };
  
  /**
   * Default enterAction which undoes/redoes to the selected node.
   */
  private Action enterAction =
          new AbstractAction("enterAction") {
    public void actionPerformed(ActionEvent evt) {
      UndoRedoManager.this.processUndoRedoCommand();
    }
  };
  
  
  /**
   * Go through the vector and cancel (remove) redo/undo adjacent pairs
   * <br>Input arg is unchanged.
   * @param fullPath the complete list of nodes in the path from the top of the tree to the selected node
   * @return the compressed list with paired undo/redo nodes and BranchUndoRedo nodes removed. This input list is unchanged.
   */
  protected Vector<UndoRedoTreeNode> compressFullPath(Vector<UndoRedoTreeNode> fullPath) {
    UndoRedoTreeNode lastNode = null;
    if (fullPath.size() > 0) {
      // keep the node we want to undo to to make sure it is not removed.
      lastNode = fullPath.get(fullPath.size()-1);
    }
    Vector<UndoRedoTreeNode> compressedFullPath = new Vector<UndoRedoTreeNode>();
    for (UndoRedoTreeNode node : fullPath) {
      if (node.getUserObject() instanceof BranchUndoRedo) {
        // skip
        continue;
      } // else

      if ((compressedFullPath.size() > 0)
      && (equalsComplement(compressedFullPath.lastElement(),node))) {
        // found complement pair remove both
        compressedFullPath.removeElementAt(compressedFullPath.size()-1);
      } else {
        // just add it
         compressedFullPath.add(node);
      }
    }
    return compressedFullPath;
  }
  
  
  /**
   * Undo/redo path, skipping BranchUndoRedo nodes.
   * <br>Will work with both non-compressed and compressed paths
   * but compressed paths reduce the work number of new nodes generated.
   * <br>Checks for Dead nodes and disposes of those node below them in
   * this branch.  The test is done here because independence between
   * undos may mean some nodes below a dead node are still reachable.
   * In this simple implementation you cannot swap any nodes, so
   * all nodes below a dead node are unreachable.
   * @param fullPath the (compressed) path to be processed
   * @throws javax.swing.undo.CannotUndoException  if cannot be undone.
   * @throws javax.swing.undo.CannotRedoException  if cannot be redone.
   */
  private void undoRedoTreePath(Vector<UndoRedoTreeNode> fullPath) throws CannotUndoException, CannotRedoException {
    // first check if any are dead and set those below as unreachable
    boolean anyDead = false;
    for (UndoRedoTreeNode t: fullPath) {
      UndoRedo undoRedo = (UndoRedo)t.getUserObject();
      if (undoRedo instanceof BranchUndoRedo) {
        continue;
      }
      if (undoRedo.isDead()) {
        anyDead = true;
        continue;
      } else if (anyDead) {
        undoRedo.setUnreachable();
      }
    }
    if (anyDead) {
      // just ignore this request
      tree.repaint(); // update node display
      return;
    }
    
    // now undo all these
    // if empty just skip to end
    UndoRedoTreeNode lastNode = null;
    for (UndoRedoTreeNode t: fullPath) {
      UndoRedo undoRedo = (UndoRedo)t.getUserObject();
      if (undoRedo instanceof BranchUndoRedo) {
        continue;
      }
      
      
      if (undoRedo.isUndo()) {
        undoRedo.undo();
      } else {
        undoRedo.redo();
      }
      // add it to the top of the tree
      UndoRedo complement = undoRedo.getComplement();
      addUndoRedo(complement,false);
      lastNode = t;
      if (supportsRollBack()) {
        ICanBeRolledBack cancellableEdit = (ICanBeRolledBack)undoRedo.getUndoableEdit();
        Frame frame = new Frame(getRoot());
        // save the frame and pass the reference to the undoableEdit
        Long key = new Long(frameCounter.getAndIncrement());
        frameMarkerMap.put(key,frame);
        //System.out.println("put key:"+key+"\nframe:"+frame);
        cancellableEdit.setFrameMarker(key);
        // now add edit
      }
      if (lastNode != null) {
        tree.expandPath(new TreePath(lastNode.getPath()));          
      }

    }
    
  }
  
  
  
  
  /**
   * Look through a redo branches for the first node that can be redone.
   *<br>Recursively searches depth first and branched encountered.
   * @param fullPath returns the updated path
   * @param branch the branch to process
   * @return true is redoable found, else false if first depth first search does not find any redoable.
   *<br>In either case fullPath is updated by added all nodes scanned.  If true is returned the last node in
   * the fullPath will be the first redoable node.
   */
  private boolean processBranchForRedo(Vector<UndoRedoTreeNode> fullPath, BranchDefaultMutableTreeNode branch) {
    boolean found = false;
    // add this branch
    fullPath.add(branch);
    // look in here for first redo
    Enumeration<UndoRedoTreeNode> ce = branch.children();
    while(ce.hasMoreElements()) {
      UndoRedoTreeNode cn = ce.nextElement();
      if (!(cn instanceof BranchDefaultMutableTreeNode)) {
        fullPath.add(cn);
        if (((UndoRedo)cn.getUserObject()).canRedo()) {
          // found one we can redo
          return true;
        } // else continue
      } else {
        // tail recursion
        return processBranchForRedo(fullPath, (BranchDefaultMutableTreeNode) cn);
      }
    }
    // reach here if all this branch already redone
    return false;
  }
  
  /**
   * Find the first undo or redo node that can be undo or redone, does not check if the node is dead.
   * <br>On return fullPath will contain the path to the node found
   * This method returns nodes similar to the standard Java undo/redo manager.
   * @param fullPath should be empty on call and on return contains the node found if result is true.
   * @param forUndo true if looking for undo, false if looking for redo
   * @return true if node found, else false.
   */
  private boolean getUndoRedoPath(Vector<UndoRedoTreeNode> fullPath, boolean forUndo) {
    Enumeration<UndoRedoTreeNode> e = getRoot().children();
    if (! (e.hasMoreElements())) {
      // no children at all
      return false;
    }
    boolean found = false;
    while (e.hasMoreElements() && !found) {
      UndoRedoTreeNode n = e.nextElement();
      UndoRedo undoRedo = (UndoRedo)n.getUserObject();
      if (n instanceof BranchDefaultMutableTreeNode) {
        if (!forUndo) {
          // process redo branch
          found = processBranchForRedo(fullPath, (BranchDefaultMutableTreeNode) n);
          break;
        } else {
          // just add and keep going looking for undo
          fullPath.add(n);
        }
      } else if (undoRedo.isUndo()) {
        fullPath.add(n);
        if (forUndo) {
          // first undo Found
          found = true;
          break;
        } // else keep going as we are looking for a redo branch to process
      } else {
        // found redo on main trunk
        // should not happen
        throw new RuntimeException(" Redo found on main branch!! "+e);
      }
    }
    if (!found) {
      return false;
    }
    // else clean up path
    Vector<UndoRedoTreeNode> compressedFullPath = compressFullPath(fullPath);
    fullPath.clear();
    fullPath.addAll(compressedFullPath);
    if (fullPath.size() == 0) {
      return false;
    }
    if (fullPath.size() >1) {
      return false;
    }
    UndoRedo undoRedo = (UndoRedo)(fullPath.firstElement().getUserObject());
    if ((forUndo && undoRedo.isUndo() && undoRedo.canUndo() )
    || (!forUndo && !undoRedo.isUndo() && undoRedo.canRedo())) {
      // all undo have been undo only branches left
      return true;
    } else {
      return false;
    }
  }
  
  
  /**
   * Adds undo/redo node to the tree.
   * <br>Creates/merges Redo branches as necessary.
   * <p>
   * For Rollback frames there are three cases:
   *<br>i) A new undoableEdit is added - just keep refernce to current root at the tail so we can rollback to here
   *<br>ii) An Undo is added -  the generated RedoBranches clone down to undo as root
   * and keep refernce to rest of tree as tail.
   *<br>iii) A Redo is added -  just keep reference to the current tree at the tail so we can rollback to here.
   *
   * @param child  the undo/redo node to add.
   * @param shouldBeVisible  true if tree should display this node after add. (Not used at present.)
   */
  private void addUndoRedo(UndoRedo child, boolean shouldBeVisible) {
    // do we need to take a snapshot of the current frame
    UndoRedoTreeNode childNode =
            new UndoRedoTreeNode(child);
    
    UndoRedoTreeNode parent = getRoot();
    
    if (parent.getChildCount() <= 0) {
      // just add
      treeModel.insertNodeInto(childNode, parent, 0);
      //Make sure the user can see the lovely new node.      
      treeModel.setRoot(getRoot());  // to overcome corruption in display seems to be a V5.0 problem.
      return;
    }
    
    if (child.isUndo()) {
      // this is an Undo just add
      treeModel.insertNodeInto(childNode, parent, 0);
    } else {
      BranchDefaultMutableTreeNode branchNode = null;
      // this is a Redo command so should be able to find a matching undo in the main path
      // will either add to existing branch or create a new one
      boolean firstChildComplement = (parent.getChildCount()>0) &&
              (child.equalsComplement(((UndoRedoTreeNode)parent.getChildAt(0)).getUserObject()));
      
      boolean firstChildBranch = ((UndoRedoTreeNode)parent.getChildAt(0)).getUserObject() instanceof BranchUndoRedo;
      boolean secondChildComplement = (parent.getChildCount()>1) &&
              (child.equalsComplement(((UndoRedoTreeNode)parent.getChildAt(1)).getUserObject()));
      
      if (firstChildComplement) {
        // found undo to match this redo
        // make new branch for these
        BranchDefaultMutableTreeNode newBranch = new BranchDefaultMutableTreeNode(new BranchUndoRedo());
        branchNode = newBranch;
        treeModel.removeNodeFromParent((UndoRedoTreeNode)parent.getChildAt(0));
        treeModel.insertNodeInto(childNode,newBranch,0);
        treeModel.insertNodeInto(newBranch, parent, 0);
        // update branchUndoRedo with childNode
        
        
      } else if (firstChildBranch && secondChildComplement) {
        // absorb into first branch
        BranchDefaultMutableTreeNode topBranch = (BranchDefaultMutableTreeNode)parent.getChildAt(0);
        branchNode = topBranch;
        // remove the complement
        treeModel.removeNodeFromParent((UndoRedoTreeNode)parent.getChildAt(1));
        // add the redo to the branch
        treeModel.insertNodeInto(childNode,topBranch,0);
        
      } else {
        // need to create a new branch
        // problem is when have two redo branches and
        BranchDefaultMutableTreeNode newBranch = new BranchDefaultMutableTreeNode(new BranchUndoRedo());
        branchNode = newBranch;
        UndoRedoTreeNode topNode = (UndoRedoTreeNode)parent.getFirstChild();
        while (!(child.equalsComplement(topNode.getUserObject()))) {
          treeModel.removeNodeFromParent(topNode);
          treeModel.insertNodeInto(topNode,newBranch,newBranch.getChildCount()); // insert at end
          topNode = (UndoRedoTreeNode)parent.getFirstChild();
        }
        
        if (!(child.equalsComplement(topNode.getUserObject()))) {
          throw new RuntimeException(" Could not find complement for "+child);
        }
        treeModel.removeNodeFromParent(topNode);
        treeModel.insertNodeInto(childNode,newBranch,0);
        treeModel.insertNodeInto(newBranch, parent, 0);
      }
      branchNode.updateDisplayMessage();
    }
    //Make sure the user can see the lovely new node.
    // force redraw of tree 
    treeModel.setRoot(getRoot());  // to overcome corruption in display  seems to be a V5.0 problem
    return;
  }
  
  
  /**
   * A tree cell renderer which respects the set width.
   */
  protected class WidthEnabledCellRenderer extends DefaultTreeCellRenderer {
    
    /**
     * Overrides <code>JComponent.getPreferredSize</code> to
     * return slightly wider preferred size value.
     * and ensures at least preferred Width
     * @return preferred dimension
     */
    public Dimension getPreferredSize() {
      Dimension        retDimension = super.getPreferredSize();
      
      if(retDimension != null)
        retDimension = new Dimension(Math.max(retDimension.width + 3,preferredWidth) ,
                retDimension.height);
      return retDimension;
    }
  }
  
  
  /**
   * This class hold the information needed to
   * reset the UndoRedo tree to some prevous state
   */
  protected class Frame {
    /**
     * This is the rootTree to be restored.
     * <br>This is a shallow clone of the original tree
     */
    private UndoRedoTreeNode preRootTree;
    private UndoRedoTreeNode postRootTree;
    
    
    /**
     * Frame constructor.
     *
     * @param rootNode a shallow clone of the original tree.
     */
    protected Frame(UndoRedoTreeNode rootNode) {
      preRootTree = lastClonedTree;
      postRootTree = (UndoRedoTreeNode)rootNode.clone();
      lastClonedTree = postRootTree;
      //System.out.println(" created frame "+toString());
    }
    
    
    /**
     * Reconstucts the original tree from the rootTree and the tail
     * <br>It also resets the undo flag on each cloned node.
     * First set all branch nodes to hasBeenDone false and
     * then set all the trunk nodes to hasBeenDone true
     *
     * @return - the new root node for the undo/redo tree
     */
    public UndoRedoTreeNode getPreTreeRootNode() {
//      System.out.println(" called frame getRootNode "+
//              ((rootTree==null)?"null":rootTree.toFullString()));
      UndoRedoTreeNode rootNode = preRootTree;
      return restoreTree(rootNode);
    }

    public UndoRedoTreeNode getPostTreeRootNode() {
//      System.out.println(" called frame getRootNode "+
//              ((rootTree==null)?"null":rootTree.toFullString()));
      UndoRedoTreeNode rootNode = postRootTree;
      return restoreTree(rootNode);
    }
    
    private UndoRedoTreeNode restoreTree(UndoRedoTreeNode rootNode) {  
      // update the branches to hasBeenDone == false
      Enumeration<UndoRedoTreeNode> e = rootNode.children();
      while(e.hasMoreElements()) {
        UndoRedoTreeNode node = e.nextElement();
        if (node.getUserObject() instanceof BranchUndoRedo) {
          restoreBranchNodes(node);
        } // continue
      }
      // then update the trunk to hasBeenDone == true
      restoreTrunkNodes(rootNode);
//      System.out.println(" frame returning "+rootNode.toFullString());
//      if (rootNode != null){
//        Enumeration estr =  rootNode.breadthFirstEnumeration();
//        while (estr.hasMoreElements()) {
//          System.out.println(estr.nextElement().toString()+"\n");
//        }
//      } else {
//        System.out.println("null rootTree \n");
//      }

      return rootNode;
    }
    
    /**
     * Sets trunk nodes to have been done.
     *
     * @param node the rootNode of tree
     */
    private void restoreTrunkNodes(UndoRedoTreeNode rootNode) {
      Enumeration<UndoRedoTreeNode> e = rootNode.children();
      while(e.hasMoreElements()) {
        UndoRedoTreeNode node = e.nextElement();
        if (node.getUserObject() instanceof BranchUndoRedo) {
          // skip as this has already been done
        } else {
          UndoRedo undoRedo = (UndoRedo)node.getUserObject();
          if (undoRedo.getUndoableEdit() instanceof ICanBeRolledBack) {
            // redo node are not done yet
            if (undoRedo.getUndoableEdit().canRedo()) {
              ((ICanBeRolledBack)(undoRedo.getUndoableEdit())).setUndoState(true);
            }
          }
        }
      }
    }
    
    /**
     * Sets branch node UndoableEdits to not done.
     *
     * @param node the Branch node to set
     */
    private void restoreBranchNodes(UndoRedoTreeNode node) {
      if (node.getUserObject() instanceof BranchUndoRedo) {
        // restore recursively
        Enumeration<UndoRedoTreeNode> e = node.children();
        while(e.hasMoreElements()) {
          // add clones of children to rtnTree
          restoreBranchNodes(e.nextElement());
        }
      } else {
        UndoRedo undoRedo = (UndoRedo)node.getUserObject();
        if (undoRedo.getUndoableEdit() instanceof ICanBeRolledBack) {
          // redo node are not done yet
          if (undoRedo.getUndoableEdit().canUndo()) {
            ((ICanBeRolledBack)(undoRedo.getUndoableEdit())).setUndoState(false);
          }
        }
      }
    }
    
    /**
     * String representation of the postTree for this Frame
     * @return the postTree for this frame
     */
    public String toString() {
      String str = "Frame rootTree:"+((postRootTree==null)?"null":postRootTree.toFullString())+"\n";
      return str;
    }
  }
  
  protected class UndoRedoTreeSelectionListener implements TreeSelectionListener {
    
    /**
     * Reference to the previously selected node's IHaveDescription class,
     * <br>may be null.
     */
    private IHaveDescription previousHaveDescription = null;
    
    /**
     * Reference to the previously selected node's IDispalyText wrapper,
     * <br>may be null.
     */
    private TextAreaWrapper previousWrapper = null;
    
    public UndoRedoTreeSelectionListener() {      
    }
    
    private IHaveDescription getIHaveDescription(UndoRedoTreeNode lastNode) {
      if (lastNode instanceof BranchDefaultMutableTreeNode) {
        return (IHaveDescription)lastNode;
      } else {
        UndoableEdit edit = lastNode.getUndoableEdit();
        if (edit instanceof IHaveDescription) {
          return (IHaveDescription)edit;
        } // else
        return null;
      }  
    }
            
    /**
     * 
     * @param e 
     */
    public void valueChanged(TreeSelectionEvent e) {
      UndoRedoTreeNode lastNode = (UndoRedoTreeNode)e.getPath().getLastPathComponent();
      IHaveDescription haveDescription = getIHaveDescription(lastNode);
      if (e.isAddedPath()) {
        // added -> ask for description
        // only seem to get these for single selection
        if ((previousHaveDescription != null) && 
            (previousHaveDescription != haveDescription)) {
            previousHaveDescription.cancelGetDescription();
        }    
        if (previousWrapper != null) {
          previousWrapper.cancel();
        }
        getTextDisplayComponent().setText(""); // clear text always
        TextAreaWrapper wrapper = null;
        if (haveDescription != null) {
          wrapper = new TextAreaWrapper(getTextDisplayComponent());
          haveDescription.getDescription(wrapper, lastNode);
        }  
        previousHaveDescription = haveDescription; // may be null
        previousWrapper = wrapper; // may be null
      } else {
        // remove -> cancel description
        if ((previousHaveDescription != null)){ 
            previousHaveDescription.cancelGetDescription();
        }    
        if (previousWrapper != null) {
          previousWrapper.cancel();
        }
        getTextDisplayComponent().setText(""); // clear text always
      }
    }  
  }

  protected class TextAreaWrapper implements IDisplayText {
    private IDisplayText displayArea;
    /**
     * 
     * @param displayArea 
     */
    public TextAreaWrapper(IDisplayText displayArea) {
      this.displayArea = displayArea;
    }
    /**
     * 
     * @param text 
     */
    public synchronized void setText(String text) {
      if (displayArea != null) {
        displayArea.setText(text);
      }
    }
    public synchronized void cancel() {
      displayArea = null;
    }
  }
  
  protected class DisplayTextArea extends JTextArea implements IDisplayText {
    public DisplayTextArea() {
      super();
      setEditable(false);
      setFocusable(false);
    }
  }
  
  

  /**
   * Get the Component used to display extended descriptions
   * @return The IDisplayText component which extends Component.
   */
  public IDisplayText getTextDisplayComponent() {
    return textDisplayComponent;
  }

  /**
   * Sets the Component used to display extended descriptions.
   *
   * @param textDisplayComponent The Component that will display extended descriptions.
   */
  public void setTextDisplayComponent(IDisplayText textDisplayComponent) {
   if (!(textDisplayComponent instanceof Component)) {
      throw new IllegalArgumentException("textDisplayComponent must extend from Component.");
    }
    this.textDisplayComponent = textDisplayComponent;
    textScrollPane.setViewportView((Component)this.textDisplayComponent);
  }

  
}

